#app{font-family:Avenir,Helvetica,Arial,sans-serif;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;text-align:center;color:#b7b7b7}#nav{padding:5px}#nav a{text-decoration:none;font-size:35px;color:#457}#nav a.router-link-exact-active{color:#42b983}h4{font-size:20px}button{filter:invert(1)}body{background:#000}a{color:#79b}ul[data-v-d6d34804]{list-style-type:none}li[data-v-d6d34804]{margin:10px}img.loading[data-v-85bcbca0]{width:30px;height:auto;position:relative;bottom:-3px}.is_drilled[data-v-85bcbca0]{border:2px solid red}.swank[data-v-2aef0fc7]{border:1px solid #39495c;margin-bottom:18px;background:#000;transition:all .5s linear;overflow-x:hidden;overflow-y:auto;height:100%}.swank.is_active[data-v-2aef0fc7]{flex:1;margin-left:10px;margin-right:10px}.swank:not(.is_active)>[data-v-2aef0fc7]{display:none}.swank.is_child_descendant[data-v-2aef0fc7],.swank.is_parent_ancestor[data-v-2aef0fc7]{opacity:0;flex:0.0001}.swank.is_child_and_button[data-v-2aef0fc7],.swank.is_parent_and_button[data-v-2aef0fc7]{cursor:pointer;opacity:.6;flex:0.01 30px}.swank.is_child_and_button[data-v-2aef0fc7],.swank.is_child_descendant[data-v-2aef0fc7],.swank.is_parent_ancestor[data-v-2aef0fc7],.swank.is_parent_and_button[data-v-2aef0fc7]{width:30px}table[data-v-2aef0fc7]{text-align:left}td[data-v-2aef0fc7]:first-child{font-family:consolas,Liberation Mono,courier,monospace}a.favorite[data-v-2aef0fc7]{cursor:pointer}a.favorite img[data-v-2aef0fc7]{width:30px;height:30px}table.top-bar[data-v-2aef0fc7]{width:100%}.swank-trail[data-v-a29b1de6]{display:flex;flex-direction:row;align-items:center;position:absolute;top:80px;bottom:80px;left:0;right:0}