# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''A package for exposing the GridRoyale frontend interface as a local web server.'''

from .server import *